var init = function () {
    var modules = {};
    var SamsungMraid = modules.SamsungMraid = typeof samsungMraid == 'undefined' ? {} : samsungMraid;
    var mraid = modules.mraid = {};
    mraid.VERSION = '2.0';

    var STATE = {
        'LOADING': 'loading',
        'DEFAULT': 'default',
        'EXPANDED': 'expanded',
        'RESIZED': 'resized',
        'HIDDEN': 'hidden',
    };
    var currentState = STATE.LOADING;
    var currentVisibility = false;
    var placementType = 'unknown';
    var supportedFeature = new Map([
        ['sms', false],
        ['tel', false],
        ['calendar', false],
        ['storePicture', false],
        ['inlineVideo', false],
    ]);
    var EVENT = {
        'READY': 'ready',
        'ERROR': 'error',
        'STATE_CHANGE': 'stateChange',
        'VIEWABLE_CHANGE': 'viewableChange',
        'SIZE_CHANGE': 'sizeChange',
    }
    var eventListenerMap = new Map();

    var orientationProperties = {
        'allowOrientationChange': true,
        'forceOrientation': 'none',
    };
    var currentPosition = {};
    var defaultPosition = {};
    var screenSize = {};
    var maxSize = {};
    var lastSizeChange = { 'width': 0, 'height': 0 };

    // isModal: READ-ONLY. always true in MRAID 2.0
    var expandProperties = {
        'useCustomClose': false,
        'isModal': true,
    };
    var resizeProperties = {
        'customClosePosition': 'top-right',
        'allowOffscreen': true,
    };
    var CUSTOM_CLOSE_POSITION = new Set(['top-left', 'top-right', 'center',
        'bottom-left', 'bottom-right', 'top-center', 'bottom-center']);

    mraid.getState = function () {
        return currentState;
    };

    mraid.getVersion = function () {
        return mraid.VERSION;
    };

    mraid.open = function (url) {
        if (url == undefined) {
            fireEvent(EVENT.ERROR, 'parameter url is mandatory', 'open');
            return;
        }
        if (typeof url !== 'string') {
            fireEvent(EVENT.ERROR, 'type of parameter url should be string', 'open');
            return;
        }
        callNative('open', url);
    };

    mraid.close = function () {
        if (mraid.getState() === STATE.HIDDEN) {
            fireEvent(EVENT.ERROR, 'close can not be applied if state is already hidden', 'close');
        } else {
            callNative('close');
        }
    };

    mraid.getPlacementType = function () {
        return placementType;
    };

    mraid.addEventListener = function (event, listener) {
        if (event == undefined || listener == undefined) {
            fireEvent(EVENT.ERROR, 'event and listener should be present',
                'addEventListener');
            return;
        }
        if (!Object.values(EVENT).includes(event)) {
            fireEvent(EVENT.ERROR,
                'event ' + event + ' is not defined in mraid 2.0',
                'addEventListener');
            return;
        }

        if (!eventListenerMap.has(event)) {
            eventListenerMap.set(event, new Set());
        }
        eventListenerMap.get(event).add(listener);
    };

    mraid.removeEventListener = function (event, listener) {
        if (event == undefined) {
            fireEvent(EVENT.ERROR, 'event should be present',
                'removeEventListener');
            return;
        }
        if (!Object.values(EVENT).includes(event)) {
            fireEvent(EVENT.ERROR,
                'event ' + event + ' is not defined in mraid 2.0',
                'removeEventListener');
            return;
        }
        if (eventListenerMap.has(event)) {
            if (listener == undefined) {
                // if listener is not specified,
                // remove all listeners for the event per MRAID spec
                eventListenerMap.delete(event);
            } else {
                var listenerSet = eventListenerMap.get(event);
                listenerSet.delete(listener);
            }
        }
    };

    mraid.supports = function (feature) {
        if (!supportedFeature.has(feature)) {
            fireEvent(EVENT.ERROR, 'feature ' + feature + ' is not defined', 'supports');
            return false;
        }
        return supportedFeature.get(feature);
    };
    mraid.getOrientationProperties = function () {
        return orientationProperties;
    };

    mraid.setOrientationProperties = function (props) {
        if (props == undefined) {
            fireEvent(EVENT.ERROR, 'argument should be present', 'setOrientationProperties');
            return;
        }
        if (props.hasOwnProperty('allowOrientationChange')) {
            if (typeof props.allowOrientationChange === 'boolean') {
                orientationProperties.allowOrientationChange =
                    props.allowOrientationChange;
            } else {
                fireEvent(EVENT.ERROR,
                    'allowOrientationChange should be typeof boolean',
                    'setOrientationProperties');
            }
        }
        if (props.hasOwnProperty('forceOrientation')) {
            if (props.forceOrientation === 'none' ||
                props.forceOrientation === 'portrait' ||
                props.forceOrientation === 'landscape') {
                orientationProperties.forceOrientation = props.forceOrientation;
            } else {
                fireEvent(EVENT.ERROR,
                    'forceOrientation should be one of none,portrait,landscape',
                    'setOrientationProperties');
            }
        }
        callNative('setOrientationProperty',
            orientationProperties.allowOrientationChange,
            orientationProperties.forceOrientation);
    };
    mraid.getCurrentPosition = function () {
        return {
            x: currentPosition.x,
            y: currentPosition.y,
            width: currentPosition.width,
            height: currentPosition.height
        };
    };
    mraid.getDefaultPosition = function () {
        return {
            x: defaultPosition.x,
            y: defaultPosition.y,
            width: defaultPosition.width,
            height: defaultPosition.height
        };
    };
    mraid.getScreenSize = function () {
        return {
            width: screenSize.width,
            height: screenSize.height
        };
    };
    mraid.getMaxSize = function () {
        return {
            width: maxSize.width,
            height: maxSize.height
        };
    };

    mraid.expand = function (url) {
        if (!(mraid.getState() === STATE.DEFAULT || mraid.getState() === STATE.RESIZED)) {
            fireEvent(EVENT.ERROR,
                'expand() only can be applied in default or resized state',
                'expand');
            return;
        }
        if (url == undefined) {
            callNative('expand', '', expandProperties.useCustomClose);
        } else {
            callNative('expand', url, expandProperties.useCustomClose);
        }
    };
    mraid.setExpandProperties = function (props) {
        if (props == undefined) {
            fireEvent(EVENT.ERROR, 'argument should be present', 'setExpandProperties');
            return;
        }
        if (props.hasOwnProperty('width')) {
            if (Number.isSafeInteger(props.width)) {
                expandProperties.width = props.width;
            } else {
                fireEvent(EVENT.ERROR, 'width should be Number.isSafeInteger() ' +
                props.width, 'setExpandProperties');
            }
        }
        if (props.hasOwnProperty('height')) {
            if (Number.isSafeInteger(props.height)) {
                expandProperties.height = props.height;
            } else {
                fireEvent(EVENT.ERROR, 'height should be Number.isSafeInteger() ' +
                props.height, 'setExpandProperties');
            }
        }
        if (props.hasOwnProperty('useCustomClose')) {
            if (typeof props.useCustomClose === 'boolean') {
                expandProperties.useCustomClose = props.useCustomClose;
            } else {
                fireEvent(EVENT.ERROR, 'useCustomClose should be typeof boolean ' +
                props.useCustomClose, 'setExpandProperties');
            }
        }
    };

    mraid.getExpandProperties = function () {
        return {
            'width': expandProperties.width,
            'height': expandProperties.height,
            'useCustomClose': expandProperties.useCustomClose,
            'isModal': expandProperties.isModal,
        };
    };

    mraid.resize = function () {
        if (!(mraid.getState() === STATE.DEFAULT ||
            mraid.getState() === STATE.RESIZED)) {
            fireEvent(EVENT.ERROR,
                'resize() only can be applied in default or resized state',
                'resize');
            return;
        }
        if (!(resizeProperties.hasOwnProperty('width') &&
            resizeProperties.hasOwnProperty('height') &&
            resizeProperties.hasOwnProperty('offsetX') &&
            resizeProperties.hasOwnProperty('offsetY'))) {

            fireEvent(EVENT.ERROR,
                'creatives must setResizeProperties BEFORE attempting to resize()',
                'resize');
            return;
        }
        callNative('resize', JSON.stringify(resizeProperties));
    };

    mraid.setResizeProperties = function (props) {
        if (props == undefined) {
            fireEvent(EVENT.ERROR, 'argument should be present', 'setResizeProperties');
            return;
        }
        if (props.hasOwnProperty('width')) {
            if (Number.isSafeInteger(props.width)) {
                resizeProperties.width = props.width;
            } else {
                fireEvent(EVENT.ERROR, 'width should be Number.isSafeInteger() ' +
                    props.width, 'setResizeProperties');
            }
        }
        if (props.hasOwnProperty('height')) {
            if (Number.isSafeInteger(props.height)) {
                resizeProperties.height = props.height;
            } else {
                fireEvent(EVENT.ERROR, 'height should be Number.isSafeInteger() ' +
                    props.height, 'setResizeProperties');
            }
        }
        if (props.hasOwnProperty('offsetX')) {
            if (Number.isSafeInteger(props.offsetX)) {
                resizeProperties.offsetX = props.offsetX;
            } else {
                fireEvent(EVENT.ERROR, 'offsetX should be Number.isSafeInteger() ' +
                    props.offsetX, 'setResizeProperties');
            }
        }
        if (props.hasOwnProperty('offsetY')) {
            if (Number.isSafeInteger(props.offsetY)) {
                resizeProperties.offsetY = props.offsetY;
            } else {
                fireEvent(EVENT.ERROR, 'offsetY should be Number.isSafeInteger() ' +
                    props.offsetY, 'setResizeProperties');
            }
        }
        if (props.hasOwnProperty('customClosePosition')) {
            if (!CUSTOM_CLOSE_POSITION.has(props.customClosePosition)) {
                fireEvent(EVENT.ERROR, 'customClosePosition should be one of ' +
                    JSON.stringify(Array.from(CUSTOM_CLOSE_POSITION.values())),
                    'setResizeProperties');
            } else {
                resizeProperties.customClosePosition = props.customClosePosition;
            }
        }
        if (props.hasOwnProperty('allowOffscreen')) {
            if (typeof props.allowOffscreen === 'boolean') {
                resizeProperties.allowOffscreen = props.allowOffscreen;
            } else {
                fireEvent(EVENT.ERROR, 'allowOffscreen should be typeof boolean ' +
                    props.allowOffscreen, 'setResizeProperties');
            }
        }
    };
    mraid.getResizeProperties = function () {
        return {
            'width': resizeProperties.width,
            'height': resizeProperties.height,
            'offsetX': resizeProperties.offsetX,
            'offsetY': resizeProperties.offsetY,
            'customClosePosition': resizeProperties.customClosePosition,
            'allowOffscreen': resizeProperties.allowOffscreen,
        };
    };
    mraid.isViewable = function () {
        return currentVisibility;
    };

    mraid.playVideo = function (url) {
        if (url == undefined) {
            fireEvent(EVENT.ERROR, 'parameter url is mandatory', 'playVideo');
            return;
        }
        if (typeof url !== 'string') {
            fireEvent(EVENT.ERROR, 'type of parameter url should be string', 'playVideo');
            return;
        }
        callNative('playVideo', url);
    };
    mraid.storePicture = function (url) {
        if (url == undefined) {
            fireEvent(EVENT.ERROR, 'parameter url is mandatory', 'storePicture');
            return;
        }
        if (typeof url !== 'string') {
            fireEvent(EVENT.ERROR, 'type of parameter url should be string', 'storePicture');
            return;
        }
        callNative('storePicture', url);
    };
    mraid.createCalendarEvent = function (calendarEvent) {
        if (calendarEvent == undefined) {
            fireEvent(EVENT.ERROR, 'parameter calendarEvent is mandatory',
                'createCalendarEvent');
            return;
        }
        callNative('createCalendarEvent', calendarEvent);
    };
    mraid.useCustomClose = function (use) {
        if (use == undefined || typeof use !== 'boolean') {
            fireEvent(EVENT.ERROR,
                'parameter useCustomClose is mandatory & should be boolean'
                , 'useCustomClose');
            return;
        }
        expandProperties.useCustomClose = use;
        callNative('useCustomClose', use);
    };

    /////// native bridge: called from Android side //////
    var nativeBridge = modules.nativeBridge = {};
    nativeBridge.setPlacementType = function (type) {
        placementType = type;
    };
    nativeBridge.setSupportedNativeFeature = function (feature, supported) {
        supportedFeature.set(feature, supported);
    };
    nativeBridge.fireReady = function () {
        console.log("native setting finished!");
        fireEvent(EVENT.READY);
    };
    nativeBridge.fireError = function (message, action) {
        fireEvent(EVENT.ERROR, message, action);
    };
    nativeBridge.fireSizeChange = function (width, height) {
        if (width === lastSizeChange.width && height === lastSizeChange.height) {
            return;
        }
        lastSizeChange = {
            width: width,
            height: height
        };
        fireEvent(EVENT.SIZE_CHANGE, width, height);
    };
    nativeBridge.setScreenSize = function (width, height) {
        screenSize = {
            width: width,
            height: height,
        };
    };
    nativeBridge.setMaxSize = function (width, height) {
        maxSize = {
            width: width,
            height: height,
        };
        // expand properties' default width, height value is maxSize
        expandProperties.width = maxSize.width;
        expandProperties.height = maxSize.height;
    };
    nativeBridge.setCurrentPosition = function (x, y, width, height) {
        currentPosition = {
            x: x,
            y: y,
            width: width,
            height: height,
        }
    };

    var isLastCallFinished = true;
    var callQueue = [];

    // 1st: function name, after 2nd: args
    function callNative() {
        if (isLastCallFinished) {
            isLastCallFinished = false;
            fireNative.apply(null, arguments);
        } else {
            callQueue.push(arguments);
        }
        if (callQueue.length > 4) {
            console.warn(
                'native call queue is crowded. current size is ' + callQueue.length);
        }
    }

    nativeBridge.onNativeCallFinished = function () {
        if (callQueue.length > 0) {
            var arr = callQueue.pop();
            fireNative.apply(null, arr);
        } else {
            isLastCallFinished = true;
        }
    }

    function fireNative() {
        let args = [...arguments];
        var functionName = args.shift();
        switch (functionName) {
            case 'open':
                SamsungMraid.open(...args);
                break;
            case 'close':
                SamsungMraid.close();
                break;
            case 'setOrientationProperty':
                SamsungMraid.setOrientationProperty(...args);
                break;
            case 'useCustomClose':
                SamsungMraid.useCustomClose(...args);
                break;
            case 'expand':
                SamsungMraid.expand(...args);
                break;
            case 'resize':
                SamsungMraid.resize(...args);
                break;
            case 'playVideo':
                SamsungMraid.playVideo(...args);
                break;
            case 'storePicture':
                SamsungMraid.storePicture(...args);
                break;
            case 'createCalendarEvent':
                SamsungMraid.createCalendarEvent(...args);
                break;
        }
    }

    nativeBridge.setDefaultPosition = function (x, y, width, height) {
        defaultPosition = {
            x: x,
            y: y,
            width: width,
            height: height,
        }
    };
    nativeBridge.setStateChange = function (state) {
        if (currentState === state) {
            return;
        }
        currentState = state;
        fireEvent(EVENT.STATE_CHANGE, state);
    };
    nativeBridge.setViewableChange = function (visible) {
        if (currentVisibility === visible) {
            return;
        }
        currentVisibility = visible;
        fireEvent(EVENT.VIEWABLE_CHANGE, visible);
    };

    function fireEvent() {
        var event = arguments[0];
        if (!eventListenerMap.has(event)) {
            return;
        }
        var args = new Array(arguments.length - 1);
        for (var i = 1; i < arguments.length; i++) {
            args[i - 1] = arguments[i];
        }

        var callbackSet = eventListenerMap.get(event);
        callbackSet.forEach(function (entry) {
            if (typeof entry === 'string') {
                var callback = window[entry];
                if (typeof callback === 'function') {
                    callback.apply(null, args);
                }
            } else if (typeof entry === 'function') {
                entry.apply(null, args);
            }
        });
    }

    // mraidTestObj is not exposed
    var mraidTestObj = modules.mraidTestObj = {};
    mraidTestObj.getEventListenerMap = function () {
        return eventListenerMap;
    }

    return modules;
};

var load = function() {
    var modules = init();
    window.mraid = modules.mraid;
    window.nativeBridge = modules.nativeBridge;

    console.log('MRAID object loaded');

    mraid.addEventListener('error', function (message, action) {
        console.error("ERROR :" + message + " : " + action);
    });
};
